from pypdfplot._version import __version__
from pypdfplot.functions import unpack, fix_pypdf

## Legacy import for the publish() function:

## from matplotlib.pyplot import *

## Specifically, ^this^ import is the reason publish() was deprecated in v0.6.
## Use plt.savefig() instead with import pypdfplot.backend at the top of your
## script.



import argparse
from pypdfplot.functions import fix_pypdf

parser = argparse.ArgumentParser()

parser.add_argument('fname',
                    metavar = 'filename',
                    help = 'filename of the severed PyPDF file that needs fixing')

parser.add_argument(
                    '-o',
                    '--output',
                    metavar = 'filename',
                    help = ('specifies the output filename; '+
                            'if omitted the conversion will be done in place'),
                    )


##parser.add_argument('-v',
##                    '--verbose',
##                    action ='store_const',
##                    const = True, 
##                    default = False,
##                    dest ='verbose',
##                    help = 'toggles verbosity')

def main():
    args = parser.parse_args()
    fix_pypdf(args.fname,
              args.output,
              #args.verbose, #Might as well have it always on
              )


if __name__ == '__main__':
    main()



import matplotlib.pyplot as plt
from pypdfplot.classes import (PyPdfFileReader, PyPdfFileWriter,
                               PdfFileReader, b_, warnings)
from pypdfplot._version import __version__
import sys
import os
from os.path import normcase, realpath
import subprocess
import io
import pickle

__PYPDFVERSION__ = '1.0'

_packlist = []
_py_file = b_('')
_pypdf_fname = ''
_iteration = 0


if sys.version_info[0] < 3:
    input = raw_input


def available_filename(fname):
    base,ext = os.path.splitext(fname)
    i = 1
    fname = base + ext
    while os.path.isfile(fname):
        fname = base + '({:1d})'.format(i) + ext
        i += 1

    return fname


def unpack(fname = None,
            verbose = True,
            ):
    
    global _pure_py,_py_file
    _pypdf_fname = fname

    if fname == None:
        fname = os.path.basename(sys.argv[0])
    
    with open(fname,'rb') as fr:
        read_buf = fr.read()
        first1k = read_buf[:1024]
        pdf_start = first1k.find(b_("%PDF"))

        #try:
        if pdf_start >= 0:
                        
            pr = PyPdfFileReader(read_buf[pdf_start:])
            
            if verbose: print('Extracting embedded files:')
            _py_file = pr.extractEmbeddedFiles()

            if verbose: print('\nPypdfplot loaded from mixed PyPDF file')
            _pure_py = False

        #except:
        else:
                        
            fr.seek(0)
            _py_file = fr.read().replace(b_('\r\n'),b_('\n'))

            if verbose: print('\nPypdfplot loaded from Python-only file')
            _pure_py = True

    return fname


def add_page(pw,write_plot_func,**kwargs):
    plot_bytes = io.BytesIO()
    write_plot_func(plot_bytes,**kwargs)
    pr = PdfFileReader(plot_bytes)
    pw.appendPagesFromReader(pr)


def remove_file(fname,verbose = True):

    success = False
    
    if os.path.isfile(fname):
        if verbose: print('-> Removing ' + fname + '...', end = '')
        if normcase(realpath(fname)) != normcase(realpath(__file__)): 
            try:
                os.remove(fname)
                success = True
            except:
                try:
                    ## Removing files via command line sometimes helps if os.remove doesn't work:
                    del_script = "python -c \"import os, time; time.sleep(1); os.remove('{}');\"".format(fname)
                    subprocess.Popen(del_script)
                    success = True
                    
                except:
                    warnings.warn('Unable to remove ' + fname + '...!')
                    success = False
        else:
            warnings.warn('Attempt to delete __init__ file was prevented')
            success = False

        if verbose:
            if success:
                print(' Done!')
            else:
                print(' FAILED!')
    else:
        success = True

    return success


def finalize_pypdf(pw,
                   output_fname,
                   pack_list,
                   cleanup,
                   pickle_figure,
                   verbose,
                   prompt_overwrite,
                   **kwargs):

    global _py_file, _pypdf_fname
    writesuccess = False

    ## Name the output file
    if output_fname == None:
        output_fname = os.path.splitext(_pypdf_fname)[0] + '.pdf' #DO-DO: Make sure to prevent self-deletion!!
    elif os.path.splitext(output_fname)[-1] == '':
        output_fname += '.pdf'
    elif os.path.splitext(output_fname)[1] not in ['.pdf','.py']: #TO-DO: Shouldn't this be just '.pdf'?
        output_fname = os.path.splitext(output_fname)[0] + '.pdf'
        warnings.warn('Invalid extension, saving as {:s}'.format(output_fname))
    
    _py_packed_fname = output_fname[:-3] + 'py'

    ## Attach Python file and auxiliary files:
    if pickle_figure:
        if verbose: print('-> Pickling figure...')
                
        if len(pack_list):
            warnings.warn('pack_list will be ignored when pickling figure!')

        fig_fname = output_fname[:-3] + 'pkl'
        fig = plt.gcf()
        fig.canvas = plt.figure().canvas
        fdata = pickle.dumps(fig)
        pw.addAttachment(fig_fname,fdata)

        flines = [b"import pypdfplot.backend.unpack",
                  b"import matplotlib.pyplot as plt",
                  b"from pickle import load",
                  b"",
                  b"with open('" + fig_fname.encode() + b"','rb') as f:",
                  b"    fig = load(f)",
                  b"",
                  b"plt.figure(fig.number)",
                  b"",
                  b"## Plot customizations go here...",
                  b"",
                  b"plt.savefig('" + output_fname.encode() + b"',",
                  b"            pack_list = ['" + fig_fname.encode() + b"'])",
                  b"",
                  b'"""']
        
        fdata = b'\n'.join(flines) 
        pw.addAttachment(_py_packed_fname,fdata)
        
    else:
        
        for fname in pack_list:
            if verbose: print('-> Attaching '+ fname)
            with open(fname,'rb') as fa:
                fdata = fa.read()
                pw.addAttachment(fname,fdata)

        if verbose: print('-> Attaching ' + _py_packed_fname)

        fdata = _py_file + b_('\n"""')
        pw.addAttachment(_py_packed_fname,fdata)
    
    pw.setPyFile(_py_packed_fname)
    pw.setPyPDFVersion(__PYPDFVERSION__)

    ## If the output file already exists, try to remove it:
    if os.path.isfile(output_fname):
        do_overwrite = False
        if prompt_overwrite:
            warnings.warn('Local copy of ' + output_fname + ' found\nOverwrite file? (y/n)')
            yes_no = input('')
            if yes_no.strip().lower()[0] == 'y':
                do_overwrite = True
            else:
                output_fname = available_filename(output_fname)
                warnings.warn('Publishing as {:s} instead'.format(output_fname))
        else:
            do_overwrite = True

        if do_overwrite:
            try:
                os.remove(output_fname)
            except:
                warnings.warn('Unable to overwrite local file ' + output_fname)
                output_fname = available_filename(output_fname)
                warnings.warn('Publishing as {:s} instead'.format(output_fname))
    
    ## Write the output file:
    if verbose: print('\nSaving ' + output_fname + '...\n')
    with open(output_fname,'wb+') as fw:
        pw.write(fw)

    write_success = True

    ## Remove the generating python file or create a new purely Python one:
    if cleanup or not _pure_py:
        if remove_file(_pypdf_fname, verbose=verbose):
            warnings.warn(_pypdf_fname + ' removed:\nSaving script in editor will make it reappear...!\n')

    ## Write the Python file if needed, and remove it if not:
    if cleanup:
        # Remove the local copy of the packed Python file if it happens to be present
        # TO-DO: This might be overzealous, consider removing...
        remove_file(_py_packed_fname, verbose=verbose)
    else:
        if not _pure_py:
            with open(_py_packed_fname,'wb') as fw:
                fw.write(_py_file)

    ## Cleanup files if needed:
    if cleanup:
        if write_success:
            if verbose and len(pack_list): print('\nCleaning up attached files:')
            for fname in pack_list:
                remove_file(fname, verbose=verbose)
        else:
            warnings.warn("Files weren't packed into PyPDF file yet, aborting cleanup")
            

def write_pypdf(write_plot_func,
                output_fname     = None,
                pack_list       = [],
                cleanup          = True,
                multiple         = ['pickle','add_page','finalize'][0],
                force_pickle     = False,
                verbose          = True,
                prompt_overwrite = False,
                **kwargs):

    global pw, _py_file, _pypdf_fname, _iteration

    ## Init PyPpfWriter:
    if multiple == 'pickle' or _iteration == 0:
        
        if verbose: print('\nPreparing PyPDF file:')
        pw = PyPdfFileWriter()

        ## If input PyPDF file hasn't been read yet, do that now:
        if _py_file == b_(''):
            _pypdf_fname = unpack()

    ## Add a page with the plot to the PyPPDF file:
    pickle_figure = (force_pickle if multiple != 'pickle' or _iteration == 0 else True)
    if multiple != 'finalize':
        if verbose: print('Adding page...')
        add_page(pw,write_plot_func,**kwargs)

    ## Write output:
    if multiple != 'add_page':
        finalize_pypdf(pw,
                       output_fname,
                       pack_list,
                       cleanup,
                       pickle_figure,
                       verbose,
                       prompt_overwrite,
                       **kwargs)
    _iteration += 1


def fix_pypdf(input_fname,
              output_fname = None,
              verbose = True):

    if verbose: print('Fixing ' + input_fname)

    if output_fname == None:
        output_fname = input_fname
    
    pw = PyPdfFileWriter()
    temp_output = io.BytesIO()
    with open(input_fname,'rb') as fr:
        try:
            fr.seek(-1024,2)
            last1k = fr.read()
            eof_addr = last1k.rfind(b'%%EOF')
            pypdf_str = last1k[eof_addr:].split()[2]

            if pypdf_str == b'PyPDF':
                warnings.warn(input_fname + ' is already compliant PyPDF-file, skipping!')
                return
            
        except(IndexError):
            pass
        
        pr = PdfFileReader(fr)     
        pw.cloneReaderDocumentRoot(pr)
        pw.setPyPDFVersion(__PYPDFVERSION__)
        pw.write(temp_output)

    do_write = True
    if output_fname == input_fname:
        if not remove_file(input_fname, verbose=verbose):
            output_fname = _available_fname(output_fname)
            warnings.warn('File saved as ' + output_fname + 'instead')
          
    if verbose: print('Writing '+ output_fname + '...', end = '')
    temp_output.seek(0)
    with open(output_fname,'wb') as fw:
        fw.write(temp_output.read())

    if verbose: print(' Done!')




# -*- coding: utf-8 -*-
#
# vim: sw=4:expandtab:foldmethod=marker
#
# Copyright (c) 2006, Mathieu Fenniak
# Copyright (c) 2007, Ashish Kulkarni <kulkarni.ashish@gmail.com>
# Copyright (c) 2019, Dirk van den Bekerom <dcmvdbekerom@gmail.com>
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

try:
    from pypdf import PdfFileWriter,PdfFileReader
    from pypdf.generic import *
    from pypdf.utils import isString,formatWarning,PdfReadError,readUntilWhitespace
    import pypdf.utils as utils
    legacy = False
    
except(ModuleNotFoundError):
    from PyPDF4 import PdfFileWriter,PdfFileReader
    from PyPDF4.generic import *
    from PyPDF4.utils import isString,formatWarning,PdfReadError,readUntilWhitespace
    import PyPDF4.utils as utils
    legacy = True

from binascii import hexlify,unhexlify
import sys
import io
import os
import struct

def ASCIIHexEncode(self,col_width = 79):

    hexdata = hexlify(self._data) + b_('>')

    temp = b_('')
    while len(hexdata) > col_width:
        temp += hexdata[:col_width] + b_('\n')
        hexdata = hexdata[col_width:]
    temp += hexdata
    self._data = temp
 
    f = self["/Filter"]
    if isinstance(f, ArrayObject):
        f.insert(0, NameObject("/ASCIIHexDecode"))
    else:
        newf = ArrayObject()
        newf.append(NameObject("/ASCIIHexDecode"))
        newf.append(f)
        f = newf

    self[NameObject("/Filter")] = f

    try:
        # Update DecodeParms if present:
        p = self["/DecodeParms"]
        if isinstance(p, ArrayObject):
            p.insert(0, NullObject())
        else:
            newp = ArrayObject()
            newp.append(NullObject())
            newp.append(p)
            p = newp
            
        self[NameObject("/DecodeParms")] = p
        
    except(KeyError):
        pass
   
StreamObject.ASCIIHexEncode = ASCIIHexEncode


def decode(data, decodeParms=None):
    bdata = data[:-1].replace(b_('\n'),b_(''))
    return unhexlify(bdata)

if legacy:
    from PyPDF4.filters import ASCIIHexDecode
    ASCIIHexDecode.decode = staticmethod(decode)

else:
    from pypdf.filters import ASCIIHexCodec
    ASCIIHexCodec.decode = staticmethod(decode)


class PyPdfFileReader(PdfFileReader):
    
    def __init__(self, read_buf):
        ##Fix XREF table, obj length, etc.:
        in_stream = self.sanitizePDF(read_buf)

        ##From this point we have a plain old regular PDF file
        super(PyPdfFileReader,self).__init__(in_stream)

    def sanitizePDF(self,read_buf,output = None):
        first1k = read_buf[:1024]
        py_obj = int(first1k.split()[1])

        last1k = read_buf[-1024:]
        startxref_addr = last1k.rfind(b_('startxref'))+len(b_('startxref\n'))
        filesize_addr = last1k.rfind(b_('%%EOF'))+len(b_('%%EOF\n'))

        old_size = int(last1k[filesize_addr:].split()[0])
        offset = len(read_buf) - old_size + 1 
        startxref = int(last1k[startxref_addr:].split()[0]) + offset

        br = io.BytesIO(read_buf)

        # Update object length:
        br.seek(0,0)
        line = br.readline()
        i1 = line.rfind(b_('/Length'))+len(b_('/Length'))
        i2 = line.rfind(b_('>>'))
        length_len = i2-i1
        br.seek(i1)
        obj_length = int(br.read(length_len))
        new_length = obj_length + offset
        br.seek(i1)
        length_str = b_((' {:'+'{:d}'.format(length_len-2)+'d} ').format(new_length)) 
        br.write(length_str)

        #Update xref table:
        br.seek(startxref,0)
        br.readline()
        nobj = int(br.readline().split()[1])
        br.readline()

        for i in range(1,nobj):
            addr = br.tell()
            obj_addr = int(br.read(10))

            if i != py_obj:
                obj_addr += offset
                br.seek(addr,0)
                br.write(b_("{:010d}".format(obj_addr)))

            br.readline()

        #Update startxref:
        br.seek(-1024,2)
        br.seek(startxref_addr,1)
        addr = br.tell()
        br.write(b_("{:d}\n%%EOF\n".format(startxref)))
        br.truncate()
        br.seek(0,0)

        ## Save to output
        if output != None:
            
            output_buf = br.read()
            with open(output,'wb') as f:
                f.write(output_buf)

        return br


    def extractEmbeddedFiles(self,verbose = True):
        
        root_obj = self._trailer['/Root']
        file_dict = root_obj['/Names']['/EmbeddedFiles']['/Names']
        
        pyname = root_obj['/PyFile']
        
        for i in range(0,len(file_dict),2):
            fname = file_dict[i]
            fobjh = file_dict[i+1]

            if isinstance(fobjh,IndirectObject):
                fobjh = fobjh.getObject()

            fobj = fobjh['/EF']['/F']

            if isinstance(fobj,IndirectObject):
                fobj = fobj.getObject()

            if fname == pyname:
                pyfile = fobj.getData()[:-4]
                if verbose: print('-> Extracting generating script: ' + fname)
                
            else:  
                if not os.path.isfile(fname):
                    fdata = fobj.getData()
                    with open(fname,'wb') as fw:
                        fw.write(fdata)
                        if verbose: print('-> Extracing ' + fname)
                else:
                    if verbose: print('-> ' + fname +' already exists, skipping')
        return pyfile


class PyPdfFileWriter(PdfFileWriter):
    def __init__(self, after_page_append=None): # in_stream,  out_stream, 

        if legacy:
            super(PyPdfFileWriter,self).__init__()
            self._stream = io.BytesIO()
            self._rootObject = self._root_object
        else:
            stream = io.BytesIO()
            super(PyPdfFileWriter,self).__init__(stream)

        '''
        Create a copy (clone) of a document from a PDF file reader

        :param reader: PDF file reader instance from which the clone
            should be created.
        :callback after_page_append (function): Callback function that is invoked after
            each page is appended to the writer. Signature includes a reference to the
            appended page (delegates to appendPagesFromReader). Callback signature:

            :param writer_pageref (PDF page reference): Reference to the page just
                appended to the document.
        '''
        debug = False
        if debug:
            print("Number of Objects: %d" % len(self._objects))
            for obj in self._objects:
                print("\tObject is %r" % obj)
                if hasattr(obj, "indirectRef") and obj.indirectRef != None:
                    print("\t\tObject's reference is %r %r, at PDF %r" % (obj.indirectRef.idnum, obj.indirectRef.generation, obj.indirectRef.pdf))

##        reader = PdfFileReader(in_stream)   
##        self.appendPagesFromReader(reader)
                
    def addAttachment(self,fname,fdata):
        ## This method fixes updating the EmbeddedFile dictionary when multiple files are attached
        try:
            old_list = self._rootObject["/Names"]["/EmbeddedFiles"]["/Names"] 
        except(KeyError):
            old_list = ArrayObject([])
        
        super(PyPdfFileWriter,self).addAttachment(fname,fdata)
        
        new_list = self._rootObject["/Names"]["/EmbeddedFiles"]["/Names"][-2:]
        if not isinstance(new_list[1],IndirectObject):
            new_list[1] = self._addObject(new_list[1])
        file_list = ArrayObject(old_list + new_list)
        self._rootObject[NameObject("/Names")][NameObject("/EmbeddedFiles")][NameObject("/Names")] = file_list
        self._rootObject[NameObject("/PageMode")] = NameObject("/UseAttachments")

    def setPyFile(self,fname):
        self._rootObject[NameObject('/PyFile')] = createStringObject(fname)

    def setPyPDFVersion(self,version):
        
        self.pypdf_version = version

        major_str, minor_str = version.split('.')[:2]
        self.major_pypdf_version = int(major_str)
        self.minor_pypdf_version = int(minor_str)
        self._rootObject[NameObject('/PyPDFVersion')] = createStringObject(version)

##    def setNewlineChar(self,newline_char):
##        self._rootObject[NameObject('/PyPDFNewlineChar')] = createStringObject(newline_char)

    def write(self,fstream):
        """
        Writes the collection of pages added to this object out as a PDF file.

        :param stream: An object to write the file to.  The object must support
            the write method and the tell method, similar to a file object.
        """
        if hasattr(self._stream, 'mode') and 'b' not in self._stream.mode:
            warnings.warn("File <%s> to write to is not in binary mode. It may not be written to correctly." % self._stream.name)
        debug = False

        if not self._root:
            self._root = self._addObject(self._rootObject)

        externalReferenceMap = {}

        # PDF objects sometimes have circular references to their /Page objects
        # inside their object tree (for example, annotations).  Those will be
        # indirect references to objects that we've recreated in this PDF.  To
        # address this problem, PageObject's store their original object
        # reference number, and we add it to the external reference map before
        # we sweep for indirect references.  This forces self-page-referencing
        # trees to reference the correct new object location, rather than
        # copying in a new copy of the page object.

        # BUT for pypdfplot we don't have to worry about this happening,
        # because we only ever get matplotlib outputs, which behave nicely.
        
##        for objIndex in range(len(self._objects)):
##            obj = self._objects[objIndex]
##            if isinstance(obj, PageObject) and obj.indirectRef != None:
##                print 'case0'
##                data = obj.indirectRef
##                if data.pdf not in externalReferenceMap:
##                    print 'case1'
##                    externalReferenceMap[data.pdf] = {}
##                if data.generation not in externalReferenceMap[data.pdf]:
##                    print 'case2'
##                    externalReferenceMap[data.pdf][data.generation] = {}
##                externalReferenceMap[data.pdf][data.generation][data.idnum] = IndirectObject(objIndex + 1, 0, self)

        self.stack = []
        if debug: print(("ERM:", externalReferenceMap, "root:", self._root))
        self._sweepIndirectReferences(externalReferenceMap, self._root)
        del self.stack

        # Begin writing:
        offsets = {}
        self._header = b_("%PDF-1.4")
        self._stream.write(b_('#') + self._header + b_(" "))

        ## Find the object number of the Python script and rearrange write order

        oi = list(range(len(self._objects)))
        try:
            pyname = self._rootObject['/PyFile']
            name_list = self._root.getObject()["/Names"]["/EmbeddedFiles"]["/Names"]
            name_dict = dict(zip(name_list[0::2],name_list[1::2]))
            py_oi = list(name_dict[pyname].getObject()['/EF'].values())[0].idnum - 1
            
            oi.pop(py_oi)
            oi = [py_oi] + oi
            
        except(KeyError):
            warnings.warn("/PyFile keyword not found, looks like a regular PDF file!")
            py_oi = -1
 
        for i in oi:
##        for i in list(range(len(self._objects))):
            idnum = i + 1
            obj = self._objects[i]
            offsets[i] = self._stream.tell()
            encryption_key = None
            
##            if hasattr(self, "_encrypt") and idnum != self._encrypt.idnum:
##                pack1 = struct.pack("<i", i + 1)[:3]
##                pack2 = struct.pack("<i", 0)[:2]
##                key = self._encrypt_key + pack1 + pack2
##                assert len(key) == (len(self._encrypt_key) + 5)
##                md5_hash = md5(key).digest()
##                key = md5_hash[:min(16, len(self._encrypt_key) + 5)]

            if i == py_oi:
                
                ##decode if necessary:
                if isinstance(obj,EncodedStreamObject):
                    obj.getData()
                    obj = obj.decodedSelf
                
                self._stream.write(b_(str(idnum) + " 0 obj "))

                obj[NameObject("/Length")] = NumberObject(len(obj._data))
                
                self._stream.write(b_("<< "))
                for key, value in list(obj.items()):
                    key.writeToStream(self._stream, encryption_key)
                    self._stream.write(b_(" "))
                    if key == '/Length':
                        space = 10 - len(str(value))
                        self._stream.write(b_(space*" "))
                    value.writeToStream(self._stream, encryption_key)
                    self._stream.write(b_(" "))
                self._stream.write(b_(">>"))

                del obj["/Length"]
                self._stream.write(b_(" stream\n"))
                data = obj._data
##                if encryption_key:
##                    data = RC4_encrypt(encryption_key, data)
                self._stream.write(data)
                self._stream.write(b_("\nendstream"))
                self._stream.write(b_("\nendobj\n"))

            else:
                self._stream.write(b_(str(idnum) + " 0 obj\n"))

                # Try to compress every object:
                if type(obj) == DecodedStreamObject:
                    obj = obj.flateEncode()

                # Hex encode object to make it compatible with Python interpreter:
                if type(obj) == EncodedStreamObject: #TO-DO: isn't this always True?
                    f = obj["/Filter"]
                    if isinstance(f, ArrayObject):
                        f = f[0]

                    if f not in ['/ASCIIHexDecode','/ASCII85Decode']:
                        #TO-DO: upgrade to /ASCII85Encode at some point
                        obj.ASCIIHexEncode()
            
                obj.writeToStream(self._stream, encryption_key)
                self._stream.write(b_("\nendobj\n"))

        # xref table
        xref_location = self._stream.tell()-1
        self._stream.write(b_("xref\n"))
        self._stream.write(b_("0 %s\n" % (len(self._objects) + 1)))
        self._stream.write(b_("%010d %05d f \n" % (0, 65535)))
        for i in range(len(self._objects)):
            self._stream.write(b_("%010d %05d n \n" % (offsets[i]-1, 0)))

        # trailer
        self._stream.write(b_("trailer\n"))
        trailer = DictionaryObject()
        trailer.update({
                NameObject("/Size"): NumberObject(len(self._objects) + 1),
                NameObject("/Root"): self._root,
                NameObject("/Info"): self._info
                })
        if hasattr(self, "_ID"):
            trailer[NameObject("/ID")] = self._ID
        if hasattr(self, "_encrypt"):
            trailer[NameObject("/Encrypt")] = self._encrypt
        trailer.writeToStream(self._stream, None)

        eof  = '\nstartxref\n{:d}\n%%EOF'.format(xref_location)
        eof += '\n{:000010d} LF\nPyPDF-' + self.pypdf_version
        eof += '\n"""\n'
        eof = b_(eof.format(self._stream.tell()+len(eof)))
        self._stream.write(eof)
            
        self._stream.seek(0)

        ## Write bytes object to fstream:
        col_width = 79
        for line in self._stream:
            while len(line) > col_width + 1:
                i = line[:col_width].rfind(b_(' '))
                fstream.write(line[:i]+b_('\n'))
                line = line[i+1:]
            fstream.write(line)






from setuptools import setup
from sys import path
path.append("pypdfplot")
from _version import __version__
import re

try:
    #If building package, complile README.rst
    files = []
    doc_folder = 'docs/source/'
    
    with open(doc_folder + 'index.rst','r') as f:
        index_text = f.read()
    files = [f+'.rst' for f in re.findall('    (\w+)',index_text)]
    
    doc = ''
    for fname in files:
        with open(doc_folder + fname,'r') as f:
           doc += f.read()
           doc += '\n'

    img_link = 'https://pypdfplot.readthedocs.io/en/latest/_images/'
    doc = doc.replace('.. image:: _static/',
                      '.. image:: ' + img_link)

    doc = re.sub(r':ref:(`[\w ]+`)', r'\1_', doc) 

    with open('README.rst','w') as f:
        f.write(doc)


##    ## Update version number in __init__ file:
##    with open('pypdfplot/__init__.py','w+b') as f:
##        buf = f.read()
##        buf[15:21] = version
##        f.seek(0)
##        f.write(buf)
    
except:
    print('Rebuilding README failed!')
    # Otherwise, just load the readme:
    with open('README.rst','r') as f:
        doc = f.read()

setup(name='pypdfplot',
    version = __version__,
    description="Saves plots as PDF with embedded generating script",
    author='Dirk van den Bekerom',
    author_email='dcmvdbekerom@gmail.com',
    license='GPLv3',
    packages=['pypdfplot'],
    install_requires=['matplotlib','PyPDF4','numpy'],
    project_urls={
        'Documentation': 'https://pypdfplot.readthedocs.io/',
        'GitHub': 'https://github.com/dcmvdbekerom/pypdfplot'}, 
    zip_safe=False,
    classifiers = [
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
         "Development Status :: 4 - Beta",
         "Intended Audience :: Science/Research",
         "Programming Language :: Python",
         "Topic :: Multimedia :: Graphics",
         "Topic :: Scientific/Engineering :: Visualization"],
    long_description=doc,
    long_description_content_type='text/x-rst',
    entry_points = {
        'console_scripts': ['fix_pypdf=pypdfplot.cli:main'],
    }
      )
